@@ -52,6 +52,15 @@ users.push({name: 'geddy'});
 users.push({name: 'neil'});
 users.push({name: 'alex'});
 
+/** Used to test code that depends on async functions being supported. */
+var testAsync = test.skip;
+try {
+  eval('(async function() {})');
+  testAsync = test;
+} catch (e) {
+  // ignore
+}
+
 suite('ejs.compile(str, options)', function () {
   test('compile to a function', function () {
     var fn = ejs.compile('<p>yay</p>');
@@ -93,6 +102,12 @@ suite('ejs.compile(str, options)', function () {
     assert.equal(fn({name: 'geddy'}), '<p>geddy</p>');
   });
 
+  test('allow customizing open and close delimiters', function() {
+    var fn;
+    fn = ejs.compile('<p>[#= name #]</p>', {delimiter: '#', openDelimiter: '[', closeDelimiter: ']'});
+    assert.equal(fn({name: 'geddy'}), '<p>geddy</p>');
+  });
+
   test('default to using ejs.delimiter', function () {
     var fn;
     ejs.delimiter = '&';
@@ -118,6 +133,76 @@ suite('ejs.compile(str, options)', function () {
     assert.equal(ejs.render(fixture('strict.ejs'), {}, {strict: true}), 'true');
   });
 
+  test('destructuring works in strict mode as an alternative to `with`', function () {
+    var locals = Object.create(null);
+    locals.foo = 'bar';
+    assert.equal(ejs.render(fixture('strict-destructuring.ejs'), locals, {
+      strict: true,
+      destructuredLocals: Object.keys(locals),
+      _with: true
+    }), locals.foo);
+  });
+
+  testAsync('destructuring works in strict and async mode', function (done) {
+    var locals = Object.create(null);
+    locals.foo = 'bar';
+    ejs.render(fixture('strict-destructuring.ejs'), locals, {
+      strict: true,
+      async: true,
+      destructuredLocals: Object.keys(locals),
+    }).then(function (value) {
+      assert.equal(value, locals.foo);
+    }).then(
+      () => done(),
+      e => done(e)
+    );
+  });
+
+  testAsync('can compile to an async function', function (done) {
+    ejs.compile('<%= await "Hi" %>', {async: true})().then(function (value) {
+      try {
+        assert.equal(value, 'Hi');
+      } catch (e) {
+        done(e);
+        return;
+      }
+
+      done();
+    });
+  });
+
+  testAsync('Non-async error message mentions `async: true`', function (done) {
+    try {
+      ejs.compile('<%= await "Hi" %>');
+    }
+    catch (err) {
+      if (err instanceof SyntaxError) {
+        assert.ok(err.message.indexOf('async: true') > -1);
+        return done();
+      } else {
+        throw err;
+      }
+    }
+    throw new Error('no error reported when there should be');
+  });
+
+  var testFuncName = typeof Object.defineProperty === 'function' ? test : test.skip;
+
+  testFuncName('Compiled function name matches `filename` without the extension', function (done) {
+    var func = ejs.compile('<%= "Foo" %>', {
+      filename: 'foo.ejs'
+    });
+
+    assert.ok(func.name === 'foo');
+    return done();
+  });
+
+  testFuncName('Compiled function name defaults to "anonymous" when `filename` is unspecified', function (done) {
+    var func = ejs.compile('<%= "Foo" %>');
+
+    assert.ok(func.name === 'anonymous');
+    return done();
+  });
 });
 
 suite('client mode', function () {
@@ -182,14 +267,14 @@ suite('client mode', function () {
 suite('ejs.render(str, dataAndOpts)', function () {
   test('render the template with data/opts passed together', function () {
     assert.equal(ejs.render('<p><?= foo ?></p>', {foo: 'yay', delimiter: '?'}),
-        '<p>yay</p>');
+      '<p>yay</p>');
   });
 
   test('disallow unsafe opts passed along in data', function () {
     assert.equal(ejs.render('<p><?= locals.foo ?></p>',
-        // localsName should not get reset because it's blacklisted
-        {_with: false, foo: 'yay', delimiter: '?', localsName: '_'}),
-        '<p>yay</p>');
+      // localsName should not get reset because it's blacklisted
+      {_with: false, foo: 'yay', delimiter: '?', localsName: '_'}),
+    '<p>yay</p>');
   });
 });
 
@@ -228,27 +313,27 @@ suite('ejs.render(str, data, opts)', function () {
 
   test('accept locals', function () {
     assert.equal(ejs.render('<p><%= name %></p>', {name: 'geddy'}),
-        '<p>geddy</p>');
+      '<p>geddy</p>');
   });
 
   test('accept locals without using with() {}', function () {
     assert.equal(ejs.render('<p><%= locals.name %></p>', {name: 'geddy'},
-                            {_with: false}),
-        '<p>geddy</p>');
+      {_with: false}),
+    '<p>geddy</p>');
     assert.throws(function() {
       ejs.render('<p><%= name %></p>', {name: 'geddy'},
-                 {_with: false});
+        {_with: false});
     }, /name is not defined/);
   });
 
   test('accept custom name for locals', function () {
     ejs.localsName = 'it';
     assert.equal(ejs.render('<p><%= it.name %></p>', {name: 'geddy'},
-                            {_with: false}),
-        '<p>geddy</p>');
+      {_with: false}),
+    '<p>geddy</p>');
     assert.throws(function() {
       ejs.render('<p><%= name %></p>', {name: 'geddy'},
-                 {_with: false});
+        {_with: false});
     }, /name is not defined/);
     ejs.localsName = 'locals';
   });
@@ -292,7 +377,7 @@ suite('ejs.render(str, data, opts)', function () {
 
 });
 
-suite('ejs.renderFile(path, [data], [options], fn)', function () {
+suite('ejs.renderFile(path, [data], [options], [fn])', function () {
   test('render a file', function(done) {
     ejs.renderFile('test/fixtures/para.ejs', function(err, html) {
       if (err) {
@@ -303,6 +388,51 @@ suite('ejs.renderFile(path, [data], [options], fn)', function () {
     });
   });
 
+  test('Promise support for reading files', function(done) {
+    var AsyncCtor;
+    var func;
+    function checkResult(html) {
+      assert.equal(html, '<p>hey</p>\n');
+    }
+    // Environments without Promise support -- should throw
+    // when no callback provided
+    function checkNoPromise() {
+      delete ejs.promiseImpl;
+      assert.throws(function () {
+        ejs.renderFile('test/fixtures/para.ejs');
+      });
+      ejs.promiseImpl = global.Promise;
+      done();
+    }
+
+    // Check for async/await support -- have to use eval for check because
+    // envs without async will break at parsing step
+    try {
+      eval('AsyncCtor = (async function () {}).constructor;');
+    }
+    catch (e) {
+      // No-op
+    }
+
+    // Both async and Promise -- in both cases, also check the call
+    // correctly throws in non-Promise envs if no callback provided
+    // -------------------
+    // Async support -- have to use eval for constructing async func for
+    // same reasons as above
+    if (AsyncCtor) {
+      func = new AsyncCtor('ejs', 'checkResult', 'checkNoPromise',
+        "let res = await ejs.renderFile('test/fixtures/para.ejs'); checkResult(res); checkNoPromise();");
+      func(ejs, checkResult, checkNoPromise);
+    }
+    // Ordinary Promise support
+    else {
+      ejs.renderFile('test/fixtures/para.ejs').then(function (res) {
+        checkResult(res);
+        checkNoPromise();
+      });
+    }
+  });
+
   test('accept locals', function(done) {
     var data =  {name: 'fonebone'};
     var options = {delimiter: '$'};
@@ -433,42 +563,6 @@ suite('ejs.renderFile(path, [data], [options], fn)', function () {
 
   });
 
-  test('support express multiple views folders, falls back to second if first is not available (include preprocessor)', function (done) {
-    var data = {
-      viewsText: 'test',
-      settings: {
-        views: [
-          path.join(__dirname, 'fixtures/nonexistent-folder'),
-          path.join(__dirname, 'fixtures')
-        ]
-      }
-    };
-    ejs.renderFile(path.join(__dirname, 'fixtures/views-old.ejs'), data, function(error, data){
-      assert.ifError(error);
-      assert.equal('<div><p>global test</p>\n</div>\n', data);
-      done();
-    });
-
-  });
-
-  test('looks relative to the containing file first (include preprocessor)', function (done) {
-    var data = {
-      viewsText: 'test',
-      settings: {
-        views: [
-          path.join(__dirname, 'fixtures/views'),
-          path.join(__dirname, 'fixtures')
-        ]
-      }
-    };
-    ejs.renderFile(path.join(__dirname, 'fixtures/views-old.ejs'), data, function(error, data){
-      assert.ifError(error);
-      assert.equal('<div><p>global test</p>\n</div>\n', data);
-      done();
-    });
-
-  });
-
   test('can reference by paths with directory names', function (done) {
     var data = {
       viewsText: 'test',
@@ -570,7 +664,7 @@ suite('cache specific', function () {
 suite('<%', function () {
   test('without semicolons', function () {
     assert.equal(ejs.render(fixture('no.semicolons.ejs')),
-        fixture('no.semicolons.html'));
+      fixture('no.semicolons.html'));
   });
 });
 
@@ -582,7 +676,7 @@ suite('<%=', function () {
 
   test('escape &amp;<script>', function () {
     assert.equal(ejs.render('<%= name %>', {name: '&nbsp;<script>'}),
-        '&amp;nbsp;&lt;script&gt;');
+      '&amp;nbsp;&lt;script&gt;');
   });
 
   test('should escape \'', function () {
@@ -611,12 +705,12 @@ suite('<%=', function () {
 suite('<%-', function () {
   test('should not throw an error with a // comment on the final line', function () {
     assert.equal(ejs.render('<%-\n// a comment\nname\n// another comment %>', {name: '&nbsp;<script>'}),
-        '&nbsp;<script>');
+      '&nbsp;<script>');
   });
 
   test('not escape', function () {
     assert.equal(ejs.render('<%- name %>', {name: '<script>'}),
-        '<script>');
+      '<script>');
   });
 
   test('terminate gracefully if no close tag is found', function () {
@@ -707,9 +801,9 @@ suite('<%%', function () {
 suite('%%>', function () {
   test('produce literal', function () {
     assert.equal(ejs.render('%%>'),
-        '%>');
+      '%>');
     assert.equal(ejs.render('  >', {}, {delimiter: ' '}),
-        ' >');
+      ' >');
   });
 });
 
@@ -822,7 +916,7 @@ suite('exceptions', function () {
 suite('rmWhitespace', function () {
   test('works', function () {
     assert.equal(ejs.render(fixture('rmWhitespace.ejs'), {}, {rmWhitespace: true}),
-        fixture('rmWhitespace.html'));
+      fixture('rmWhitespace.html'));
   });
 });
 
@@ -830,19 +924,29 @@ suite('include()', function () {
   test('include ejs', function () {
     var file = 'test/fixtures/include-simple.ejs';
     assert.equal(ejs.render(fixture('include-simple.ejs'), {}, {filename: file}),
-        fixture('include-simple.html'));
+      fixture('include-simple.html'));
   });
 
   test('include and escape ejs', function () {
     var file = 'test/fixtures/include-escaped.ejs';
     assert.equal(ejs.render(fixture('include-escaped.ejs'), {}, {filename: file}),
-        fixture('include-escaped.html'));
+      fixture('include-escaped.html'));
+  });
+
+  test('include and escape within included ejs', function () {
+    var escape = function (s) {
+      return s.toUpperCase();
+    };
+
+    var file = 'test/fixtures/include-nested-escape.ejs';
+    assert.equal(ejs.render(fixture('include-nested-escape.ejs'), {}, {filename: file, escape: escape}),
+      fixture('include-nested-escape.html'));
   });
 
   test('include in expression ejs', function () {
     var file = 'test/fixtures/include-expression.ejs';
     assert.equal(ejs.render(fixture('include-expression.ejs'), {}, {filename: file}),
-        fixture('include-expression.html'));
+      fixture('include-expression.html'));
   });
 
   test('include ejs fails without `filename`', function () {
@@ -856,6 +960,17 @@ suite('include()', function () {
     throw new Error('expected inclusion error');
   });
 
+  test('show filename when including nonexistent file', function () {
+    try {
+      ejs.render(fixture('include-nonexistent.ejs'));
+    }
+    catch (err) {
+      assert.ok(err.message.indexOf('nonexistent-file') > -1);
+      return;
+    }
+    throw new Error('expected inclusion error containing file name');
+  });
+
   test('strips BOM', function () {
     assert.equal(
       ejs.render('<%- include("fixtures/includes/bom.ejs") %>',
@@ -866,7 +981,7 @@ suite('include()', function () {
   test('include ejs with locals', function () {
     var file = 'test/fixtures/include.ejs';
     assert.equal(ejs.render(fixture('include.ejs'), {pets: users}, {filename: file, delimiter: '@'}),
-        fixture('include.html'));
+      fixture('include.html'));
   });
 
   test('include ejs with absolute path and locals', function () {
@@ -874,7 +989,7 @@ suite('include()', function () {
     assert.equal(ejs.render(fixture('include-abspath.ejs'),
       {dir: path.join(__dirname, 'fixtures'), pets: users, path: path},
       {filename: file, delimiter: '@'}),
-        fixture('include.html'));
+    fixture('include.html'));
   });
 
   test('include ejs with set root path', function () {
@@ -884,10 +999,40 @@ suite('include()', function () {
       fixture('include.html'));
   });
 
+  test('include ejs with custom includer function', function () {
+    var file = 'test/fixtures/include-root.ejs';
+    var inc = function (original, prev) {
+      if (original.charAt(0) === '/') {
+        return {
+          filename: path.join(__dirname, 'fixtures', prev)
+        };
+      } else {
+        return prev;
+      }
+    };
+    assert.equal(ejs.render(fixture('include-root.ejs'), {pets: users}, {filename: file, delimiter: '@', includer: inc}),
+      fixture('include.html'));
+  });
+
+  test('include ejs with includer returning template', function () {
+    var file = 'test/fixtures/include-root.ejs';
+    var inc = function (original, prev) {
+      if (prev === '/include.ejs') {
+        return {
+          template: '<p>Hello template!</p>\n'
+        };
+      } else {
+        return prev;
+      }
+    };
+    assert.equal(ejs.render(fixture('include-root.ejs'), {pets: users}, {filename: file, delimiter: '@', includer: inc}),
+      fixture('hello-template.html'));
+  });
+
   test('work when nested', function () {
     var file = 'test/fixtures/menu.ejs';
     assert.equal(ejs.render(fixture('menu.ejs'), {pets: users}, {filename: file}),
-        fixture('menu.html'));
+      fixture('menu.html'));
   });
 
   test('work with a variable path', function () {
@@ -900,7 +1045,7 @@ suite('include()', function () {
   test('include arbitrary files as-is', function () {
     var file = 'test/fixtures/include.css.ejs';
     assert.equal(ejs.render(fixture('include.css.ejs'), {pets: users}, {filename: file}),
-        fixture('include.css.html'));
+      fixture('include.css.html'));
   });
 
   test('pass compileDebug to include', function () {
@@ -962,135 +1107,10 @@ suite('include()', function () {
 
 });
 
-suite('preprocessor include', function () {
-  test('work', function () {
-    var file = 'test/fixtures/include_preprocessor.ejs';
-    assert.equal(ejs.render(fixture('include_preprocessor.ejs'), {pets: users}, {filename: file, delimiter: '@'}),
-        fixture('include_preprocessor.html'));
-  });
-
-  test('no false positives', function () {
-    assert.equal(ejs.render('<% %> include foo <% %>'), ' include foo ');
-  });
-
-  test('fails without `filename`', function () {
-    try {
-      ejs.render(fixture('include_preprocessor.ejs'), {pets: users}, {delimiter: '@'});
-    }
-    catch (err) {
-      assert.ok(err.message.indexOf('Could not find') > -1);
-      return;
-    }
-    throw new Error('expected inclusion error');
-  });
-
-  test('strips BOM', function () {
-    assert.equal(
-      ejs.render('<% include fixtures/includes/bom.ejs %>',
-        {}, {filename: path.join(__dirname, 'f.ejs')}),
-      '<p>This is a file with BOM.</p>\n');
-  });
-
-  test('work when nested', function () {
-    var file = 'test/fixtures/menu_preprocessor.ejs';
-    assert.equal(ejs.render(fixture('menu_preprocessor.ejs'), {pets: users}, {filename: file}),
-        fixture('menu_preprocessor.html'));
-  });
-
-  test('tracks dependency correctly', function () {
-    var file = 'test/fixtures/menu_preprocessor.ejs';
-    var fn = ejs.compile(fixture('menu_preprocessor.ejs'), {filename: file});
-    assert(fn.dependencies.length);
-  });
-
-  test('include arbitrary files as-is', function () {
-    var file = 'test/fixtures/include_preprocessor.css.ejs';
-    assert.equal(ejs.render(fixture('include_preprocessor.css.ejs'), {pets: users}, {filename: file}),
-        fixture('include_preprocessor.css.html'));
-  });
-
-  test('pass compileDebug to include', function () {
-    var file = 'test/fixtures/include_preprocessor.ejs';
-    var fn;
-    fn = ejs.compile(fixture('include_preprocessor.ejs'), {
-      filename: file,
-      delimiter: '@',
-      compileDebug: false
-    });
-    try {
-      // Render without a required variable reference
-      fn({foo: 'asdf'});
-    }
-    catch(e) {
-      assert.equal(e.message, 'pets is not defined');
-      assert.ok(!e.path);
-      return;
-    }
-    throw new Error('no error reported when there should be');
-  });
-
-  test('is static', function () {
-    fs.writeFileSync(__dirname + '/tmp/include_preprocessor.ejs', '<p>Old</p>');
-    var file = 'test/fixtures/include_preprocessor_cache.ejs';
-    var options = {filename: file};
-    var out = ejs.compile(fixture('include_preprocessor_cache.ejs'), options);
-    assert.equal(out(), '<p>Old</p>\n');
-
-    fs.writeFileSync(__dirname + '/tmp/include_preprocessor.ejs', '<p>New</p>');
-    assert.equal(out(), '<p>Old</p>\n');
-  });
-
-  test('support caching', function () {
-    fs.writeFileSync(__dirname + '/tmp/include_preprocessor.ejs', '<p>Old</p>');
-    var file = 'test/fixtures/include_preprocessor_cache.ejs';
-    var options = {cache: true, filename: file};
-    var out = ejs.render(fixture('include_preprocessor_cache.ejs'), {}, options);
-    var expected = fixture('include_preprocessor_cache.html');
-    assert.equal(out, expected);
-    fs.writeFileSync(__dirname + '/tmp/include_preprocessor.ejs', '<p>New</p>');
-    out = ejs.render(fixture('include_preprocessor_cache.ejs'), {}, options);
-    assert.equal(out, expected);
-  });
-
-  test('whitespace slurp and rmWhitespace work', function() {
-    var file = 'test/fixtures/include_preprocessor_line_slurp.ejs';
-    var template = fixture('include_preprocessor_line_slurp.ejs');
-    var expected = fixture('include_preprocessor_line_slurp.html');
-    var options = {rmWhitespace: true, filename: file};
-    assert.equal(ejs.render(template, {}, options),
-        expected);
-  });
-
-  test('handles errors in included file', function() {
-    try {
-      ejs.render('<%- include fixtures/include-with-error %>', {}, {filename: path.join(__dirname, 'f.ejs')});
-    }
-    catch (err) {
-      assert.ok(err.message.indexOf('foobar is not defined') > -1);
-      return;
-    }
-    throw new Error('expected inclusion error');
-  });
-
-});
-
 suite('comments', function () {
   test('fully render with comments removed', function () {
     assert.equal(ejs.render(fixture('comments.ejs')),
-        fixture('comments.html'));
-  });
-});
-
-suite('require', function () {
-
-  // Only works with inline/preprocessor includes
-  test('allow ejs templates to be required as node modules', function () {
-    var file = 'test/fixtures/include_preprocessor.ejs';
-    var template = require(__dirname + '/fixtures/menu_preprocessor.ejs');
-    if (!process.env.running_under_istanbul) {
-      assert.equal(template({filename: file, pets: users}),
-          fixture('menu_preprocessor.html'));
-    }
+      fixture('comments.html'));
   });
 });
 
